#!/usr/bin/python

EXPECT_STATUS = {'system/urandom': 'SERVICE_DONE',
                 'system/udev/udevd': 'DAEMON_RUNNING',
                 'system/udev': 'UP',
                 'system/udev/filldev': 'SERVICE_DONE',
                 'system/udev/set_hotplug': 'SERVICE_DONE',
                 'system/udev/mountdev': 'SERVICE_DONE',
                 'system/usb': 'SERVICE_DONE',
                 'system/swap': 'SERVICE_DONE',
                 'system/bootmisc': 'SERVICE_DONE',
                 'daemon/agetty/2': 'DAEMON_RUNNING',
                 'runtests': 'UP',
                 'net/lo': 'SERVICE_DONE',
                 'system/mountfs': 'SERVICE_DONE',
                 'system/initial': 'UP',
                 'system': 'UP',
                 'system/initial/filldev': 'SERVICE_DONE',
                 'system/hdparm': 'SERVICE_DONE',
                 'daemon/agetty': 'UP',
                 'daemon/agetty/3': 'DAEMON_RUNNING',
                 'system/hostname': 'SERVICE_DONE',
                 'system/modules': 'SERVICE_DONE',
                 'system/mountroot': 'SERVICE_DONE',
                 'system/mountroot/check': 'SERVICE_DONE',
                 'system/mountroot/rootrw': 'SERVICE_DONE',
                 'system/initial/mountvirtfs': 'SERVICE_DONE',
                 'system/test-suite': 'SERVICE_DONE',
                 'system/checkfs': 'SERVICE_DONE',
                 'system/modules/usbcore': 'SERVICE_DONE',
                 'system/modules/nvram': 'SERVICE_DONE',
                 'system/initial/loglevel': 'SERVICE_DONE',
                 'daemon/syslog-ng': 'DAEMON_RUNNING',
                 'default': 'UP'}

DAEMON_NAMES = { 'system/initial/udev': 'udevd',
                 'system/initial/udevd': 'udevd',
                 'system/udev/udevd': 'udevd',
                 'daemon/agetty/1': 'agetty',
                 'daemon/agetty/2': 'agetty',
                 'daemon/agetty/3': 'agetty',
                 'daemon/agetty/4': 'agetty',
                 'daemon/agetty/5': 'agetty',
                 'daemon/agetty/6': 'agetty',
                 'daemon/vixie-cron': 'cron',
                 'daemon/sshd': 'sshd',
                 'daemon/syslog-ng': 'syslog-ng',
                 'daemon/atd': 'atd',
                 'daemon/dbus': 'dbus-daemon'}

C_OFF = "\x1b[0m"
C_FG_BLACK = "\x1b[30m"
C_FG_RED = "\x1b[31m"
C_FG_GREEN = "\x1b[32m"
C_FG_YELLOW = "\x1b[33m"
C_FG_BLUE = "\x1b[34m"
C_FG_MAGENTA = "\x1b[35m"
C_FG_CYAN = "\x1b[36m"
C_FG_WHITE = "\x1b[37m"
C_FG_DEFAULT = "\x1b[39m"
C_FG_LIGHT_BLUE = "\x1b[34;01m"
C_FG_NEON_GREEN = "\x1b[32;01m"
C_FG_LIGHT_RED = "\x1b[01;31m"



working = { }
stats = { }
poss_results = ('PASS','FAIL','KPASS','KFAIL','SKIP')
for r in poss_results:
    stats[r] = 0

import os, sys, re, test_slave, pwd
sep_re = re.compile(r"^\s*-{10,}\s*$")
ngc_s_re = re.compile(r"^\s*(\d\d:\d\d:\d\d)\s+(\S+)\s+:\s*(\S+)(?:\s+.*$)")
bl_re = re.compile(r"^\s*$")
ngc_g_re = re.compile(r"INT_COMMAND_RESULT: (-?\d+)")
fl_re = re.compile(r"^(.*)$")
ansi_esc_re = re.compile(r"\x1b\[[\d;]+m")

def chk_worked(test):
    if not working.has_key(test):
        return None
    else:
        return working[test]

def ngc_s():
    s = { }
    p = os.popen("ngc -s","r")
    while True:
        t = ansi_esc_re.sub("",p.readline())
        if sep_re.match(t): break
        if t == '':
            print "Premature end of ngc -s output"
            return None
    while True:
        t = ansi_esc_re.sub("",p.readline())
        if t == '': break
        if bl_re.match(t):
            continue
        m = ngc_s_re.match(t)
        if m == None:
            print "Bad ngc -s line: " + t
            return None
        else:
            s[m.group(2)] = m.group(3)
    if p.close() != None:
        print "ngc -s failed"
        return None
    return s

def get_svc_state(svc):
    s = ngc_s()
    if not s.has_key(svc):
        return None
    return s[svc]

def ngc(cmd,param=None):
    """Runs ngc <cmd> <param> and returns the return code (0 on success)"""
    if param == None:
        return os.spawnlp(os.P_WAIT,"ngc","ngc",cmd)
    else:
        return os.spawnlp(os.P_WAIT,"ngc","ngc",cmd,param)

def ngc_async(cmd,param=None):
    """Runs ngc <cmd> <param> asyncronously and returns its pid"""
    if param == None:
        return os.spawnlp(os.P_NOWAIT,"ngc","ngc",cmd)
    else:
        return os.spawnlp(os.P_NOWAIT,"ngc","ngc",cmd,param)

def ngc_g(svc):
    p = os.popen("ngc -g "+svc,"r")
    pid = None
    while True:
        t = ansi_esc_re.sub("",p.readline())
        if t == '': break
        m = ngc_g_re.search(t)
        if m != None:
            pid = int(m.group(1))
    if p.close() != None:
        print "ngc -g failed"
        return None
    return pid

def do_fail(desc,kfail=False):
    do_test(desc,False,kfail)

def add_count(status):
    global stats
    stats[status] += 1

def do_test(desc,res,kfail=False):
    col = C_FG_YELLOW
    if res == None:
        status = "SKIP"; col = C_FG_YELLOW
    else:
        if res:
            status = "PASS"; col = C_FG_NEON_GREEN
        else:
            status = "FAIL"; col = C_FG_LIGHT_RED
        if kfail:
            status = "K" + status
    add_count(status)
    print C_FG_CYAN + "initng-tester: " + col + status + C_OFF + " " + desc

def do_pass(desc,kfail=False):
    do_test(desc,True,kfail)

def do_skip(desc):
    do_test(desc, None)

def do_log(msg):
    print C_FG_CYAN +"initng-tester: "+C_OFF + msg

def check_what_up(c):
    global working
    s = ngc_s(); r = { }
    if chk_worked('ngc -s') == None:
        working['ngc -s'] = (s != None)
        do_test("check ngc -s is working", (s != None))
    if not chk_worked('ngc -s'):
        do_skip("check services in correct states")
        do_skip("check for stray services")
        return
    do_log("checking statuses")
    s = ngc_s(); is_pass = True
    for it in c.iteritems():
        if s.has_key(it[0]):
            if s[it[0]] != it[1]:
                do_log(it[0] + " has status " + s[it[0]] +
                       ", expected " + it[1])
                is_pass = False
        else:
            do_log(it[0] + " has no status!")
            is_pass = False
    do_test("check services in correct states",is_pass)
    stray = 0
    for svc in s.keys():
        if not c.has_key(svc):
            if s[svc] != "STOPPED" and s[svc] != "SERVICE_STOPPED":
                do_log("unexpected service " + svc + " status " + s[svc])
                stray += 1
    do_test("check for stray services", stray == 0)

def do_sleep(t):
    os.spawnlp(os.P_WAIT,"sleep","sleep",str(t))

def chk_ngc_c():
    print "Testing ngc -c"
    if ngc("-c") != 0:
        print "Couldn't run ngc -c"
        do_fail("test 'ngc -c'")
        return
    os.spawnlp(os.P_WAIT,"sleep","sleep","10")
    if ngc_s() == None:
        print "Couldn't run ngc -s after ngc -c"
        do_fail("test 'ngc -c'")
    else:
        do_pass("test 'ngc -c'")

def chk_ngc_updown():
    global working
    do_log("testing ngc -u and ngc -d")
    s = ngc("-u","tests/daemon1")
    if s != 0:
        working['ngc -u'] = False
    do_test("check ngc -u succeeds", s==0)
    slave = test_slave.accept()
    do_test("check ngc -u actually starts daemon", slave!=None)
    working['ngc -u'] = (s == 0 and slave != None)
    if slave == None:
        do_skip("check ngc -d waits for service to stop")        
        do_skip("check ngc -s shows DAEMON_TERM") 
        do_skip("check ngc -d stops daemon")
        do_skip("check ngc -d succeeds")
        return
    pid = ngc_async("-d","tests/daemon1") # doesn't return until later 
    slave.settimeout(20.0) # seconds
    do_sleep(2)
    quitted = os.waitpid(pid,os.WNOHANG)[0] != 0
    do_test("check ngc -d waits for daemon to stop",
            not quitted)
    do_test("check ngc -s shows DAEMON_TERM",
            get_svc_state('tests/daemon1')=='DAEMON_TERM')
    success = False
    while True:
        try:
            msg = slave.get_msg()
        except test_slave.DeadSockError:
            if not success:
                do_log("Slave process died!")
            break
        if msg == None: break
        if msg[0] == test_slave.MSG_SIGTERM:
            slave.quit(0)
            success = True
    do_test("check ngc -d stops daemon",success)
    working['ngc -d'] = success
    if quitted:
        do_skip("check ngc -d succeeds")
        working['ngc -d status'] = False
    else:
        success = os.waitpid(pid,0)[1]==0 # now ngc -d should return
        do_test("check ngc -d succeeds", success)
        working['ngc -d status'] = success

def chk_service_updown():
    global working
    do_log("testing ngc -u and ngc -d on a service")
    pid = ngc_async("-u","tests/service1"); quitted = False
    slave = test_slave.accept()
    do_test("check ngc -u starts service", slave!=None)
    if slave != None:
        do_sleep(2)
        quitted = os.waitpid(pid,os.WNOHANG)[0] != 0
        do_test("check ngc -u waits for service to start",
                not quitted)
    else:
        do_skip("check ngc -u waits for service to start")
        do_skip("check ngc -s shows SERVICE_START_RUN")
        do_skip("check ngc -u succeeds")
        do_skip("check ngc -s shows SERVICE_DONE")    
        do_skip("check ngc -d stops service")
        do_skip("check ngc -d waits for service to stop")
        do_skip("check ngc -s shows SERVICE_STOP_RUN")
        do_skip("check ngc -d succeeds")
        return
    do_test("check ngc -s shows SERVICE_START_RUN",
            get_svc_state('tests/service1')=='SERVICE_START_RUN')
    slave.quit(0)
    if quitted:
        do_skip("check ngc -u succeeds")
    else:
        do_test("check ngc -u succeeds",
                os.waitpid(pid,0)[1] == 0)
    do_test("check ngc -s shows SERVICE_DONE",
            get_svc_state('tests/service1')=='SERVICE_DONE')
    pid = ngc_async("-d","tests/service1")
    slave = test_slave.accept()
    do_test("check ngc -d stops service", slave!=None)
    if slave != None:
        do_sleep(2)
        quitted = os.waitpid(pid,os.WNOHANG)[0] != 0
        do_test("check ngc -d waits for service to stop",
                not quitted)
    else:
        do_skip("check ngc -d waits for service to stop")
        return
    do_test("check ngc -s shows SERVICE_STOP_RUN",
            get_svc_state('tests/service1')=='SERVICE_STOP_RUN')
    slave.quit(1)
    if quitted or chk_worked('ngc -d status') == False:
        if not quitted:
             os.waitpid(pid,0)
        do_skip("check ngc -d fails correctly")
    else:
        do_test("check ngc -d fails correctly",
            os.waitpid(pid,0)[1] == 0)
    # should it be FAIL_STOPPING? It isn't at the moment...
    state = get_svc_state('tests/service1')
    if state != None and state != 'STOPPED':
        ngc('-z','tests/service1')

def chk_daemon_pidfile_nofork():
    global working
    do_log("testing a deamon with a pidfile which doesn't fork")
    pid = ngc_async("-u","tests/daemon-nofork"); quitted = False
    slave = test_slave.accept()
    do_test("check ngc -u starts daemon", slave!=None)
    if slave != None:
        do_sleep(2)
        quitted = os.waitpid(pid,os.WNOHANG)[0] != 0
        do_test("check ngc -u waits for daemon to start",
                not quitted)
    else:
        do_skip("check ngc -u waits for daemon to start")
        do_skip("check ngc -s shows DAEMON_WAIT_FOR_PID_FILE")
        do_skip("check ngc -u succeeds")
        do_skip("check ngc -s shows DAEMON_RUNNING")
        do_skip("check ngc -d stops daemon")
        do_skip("check ngc -d succeeds")
        return
    do_test("check ngc -s shows DAEMON_WAIT_FOR_PID_FILE",
            get_svc_state('tests/daemon-nofork')=='DAEMON_WAIT_FOR_PID_FILE')
    f = open("/var/run/test-daemon-nofork.pid","w")
    f.write(str(slave.get_pinfo()[0])); f.close()
    if quitted:
        do_skip("check ngc -u succeeds")
    else:
        do_test("check ngc -u succeeds",
                os.waitpid(pid,0)[1] == 0)
    do_test("check ngc -s shows DAEMON_RUNNING",
            get_svc_state('tests/daemon-nofork')=='DAEMON_RUNNING')
    pid = ngc_async("-d","tests/daemon-nofork") # doesn't return until later
    slave.settimeout(20.0) # seconds
    success = False
    while True:
        try:
            msg = slave.get_msg()
        except test_slave.DeadSockError:
            if not success:
                do_log("Slave process died!")
            break
        if msg == None: break
        if msg[0] == test_slave.MSG_SIGTERM:
            slave.quit(0)
            success = True
    do_test("check ngc -d stops daemon",success)
    success = os.waitpid(pid,0)[1]==0 # now ngc -d should return
    if working['ngc -d status']:
        do_test("check ngc -d succeeds", success)
    else:
        do_skip("check ngc -d succeeds")

def chk_daemon_pidfile_fork():
    global working
    do_log("testing a deamon with a pidfile which forks")
    pid = ngc_async("-u","tests/daemon-fork"); quitted = False
    slave = test_slave.accept()
    do_test("check ngc -u starts daemon", slave!=None)
    if slave != None:
        do_sleep(2)
        quitted = os.waitpid(pid,os.WNOHANG)[0] != 0
        do_test("check ngc -u waits for daemon to start",
                not quitted)
    else:
        do_skip("check ngc -u waits for daemon to start")
        do_skip("check ngc -s shows DAEMON_LAUNCH")
        do_skip("check ngc -s shows DAEMON_WAIT_FOR_PID_FILE")
        do_skip("check ngc -u succeeds")
        do_skip("check ngc -s shows DAEMON_RUNNING")
        do_skip("check ngc -d stops daemon")
        do_skip("check ngc -d succeeds")
        return
    do_test("check ngc -s shows DAEMON_LAUNCH",
            get_svc_state('tests/daemon-fork')=='DAEMON_LAUNCH')
    slave.fork_and_ret(); do_sleep(2)
    do_test("check ngc -s shows DAEMON_WAIT_FOR_PID_FILE",
            get_svc_state('tests/daemon-fork')=='DAEMON_WAIT_FOR_PID_FILE')
    f = open("/var/run/test-daemon-nofork.pid","w")
    f.write(str(slave.get_pinfo()[0])); f.close()
    if quitted:
        do_skip("check ngc -u succeeds")
    else:
        do_test("check ngc -u succeeds",
                os.waitpid(pid,0)[1] == 0)
    do_test("check ngc -s shows DAEMON_RUNNING",
            get_svc_state('tests/daemon-fork')=='DAEMON_RUNNING')
    pid = ngc_async("-d","tests/daemon-fork") # doesn't return until later
    slave.settimeout(20.0) # seconds
    success = False
    while True:
        try:
            msg = slave.get_msg()
        except test_slave.DeadSockError:
            if not success:
                do_log("Slave process died!")
            break
        if msg == None: break
        if msg[0] == test_slave.MSG_SIGTERM:
            slave.quit(0)
            success = True
    do_test("check ngc -d stops daemon",success)
    success = os.waitpid(pid,0)[1]==0 # now ngc -d should return
    if working['ngc -d status']:
        do_test("check ngc -d succeeds", success)
    else:
        do_skip("check ngc -d succeeds")


def chk_all_types(name, callback,svcsuf=""):
    do_log("checking "+name+" works")
    if callback == None or not (chk_worked('ngc -u') and chk_worked('ngc -d')):
        do_skip("checking "+name+" works for exec start")
        do_skip("checking "+name+" works for exec stop")
        do_skip("checking "+name+" works for exec daemon")
        do_skip("checking "+name+" works for exec kill")
        return
    svc = "tests/test-"+name+"-service"+svcsuf
    pid = ngc_async("-u",svc)
    slave = test_slave.accept()
    
    if slave == None:
        do_log("exec start didn't give me a test_slave!")
        do_fail("checking "+name+" works for exec start")
        do_skip("checking "+name+" works for exec stop")
        os.waitpid(pid,0)
    else:
        do_test("checking "+name+" works for exec start",callback(slave))
        slave.quit(0)
        os.waitpid(pid,0)
        pid = ngc_async("-d",svc)
        slave = test_slave.accept()
        if slave == None:
            do_log("exec stop didn't give me a test_slave!")
            do_fail("checking "+name+" works for exec stop")
        else:
            do_test("checking "+name+" works for exec stop",callback(slave))
            slave.quit(0)
        os.waitpid(pid,0)
    svc = "tests/test-"+name+"-daemon"+svcsuf
    ngc("-u",svc)
    slave = test_slave.accept()
    
    if slave == None:
        do_log("exec daemon didn't give me a test_slave!")
        do_fail("checking "+name+" works for exec daemon")
        do_skip("checking "+name+" works for exec kill")
    else:
        do_test("checking "+name+" works for exec daemon",callback(slave))
        pid = ngc_async("-d",svc)
        kslave = test_slave.accept()
        if kslave == None:
            do_log("exec stop didn't give me a test_slave!")
            do_fail("checking "+name+" works for exec kill")
            slave.quit(0)
        else:
            do_test("checking "+name+" works for exec kill",callback(kslave))
            slave.quit(0)
            kslave.quit(0)
        os.waitpid(pid,0)

def _chk_chdir_cb(slave):
    return slave.get_cwd() == "/usr/lib/initng-test/testdir"

def _chk_env_cb(slave):
    works = True
    t = slave.get_env("WOMBAT")
    if t != "fed":
        do_log("Got WOMBAT=" + t); works = False
    t = slave.get_env("FOO")
    if t != "123456":
        do_log("Got FOO=" + t); works = False
    t = slave.get_env("BAR")
    if t != "baz":
        do_log("Got BAR=" + t); works = False 
    t = slave.get_env("OPTIONS")
    if t != "-s /var/run/acpid.socket":
        do_log("Got OPTIONS=" + t); works = False    
    t = slave.get_env("MODULES")
    if t != "battery ac processor button fan thermal":
        do_log("Got MODULES=" + t); works = False
    t = slave.get_env("HOME")
    if t != "/usr/lib/initng-test":
        do_log("Got HOME=" + t); works = False
#    t = slave.get_env("PATH")
#    if t == None or t.find("/root/of/all/evil") < 0:
#        do_log("Got PATH=" + t); works = False
    return works

def _chk_suid_cb(uid,slave):
    pinfo = slave.get_pinfo()
    if pinfo == None:
        do_log("MSG_GET_PINFO failed!")
        return None
    else:
        return pinfo[1] == uid

def summary():
    print "*** initng-tester SUMMARY ***"
    for s in poss_results:
        print (s+":").ljust(8) + " "  + str(stats[s])
    print "*****************************"

def get_name_by_pid(pid):
    s = { }
    p = os.popen("ps -p " + str(pid) + " -o comm=")
    t = p.readline()
    m = fl_re.match(t)
    if t == "":
        return None
    elif m == None:
        return None
    elif m.group(1) == "":
        return None
    else:
        return m.group(1)


def chk_really_runs():
    svcs = ngc_s()
    print "Checking everything is really running"
    if svcs == None:
        print "Running ngc -s failed!"
        do_fail("check daemons are really running")
        return
    for svc in svcs.keys():
        if DAEMON_NAMES.has_key(svc):
            exe_name = DAEMON_NAMES[svc]
        else:
            exe_name = None
        if svcs[svc] == 'DAEMON_RUNNING':
            pid = ngc_g(svc)
            if pid == None:
                print "'ngc -g " + svc + "' failed!"
                do_fail("check " + svc + " is running")
                continue
            elif pid < 1:
                print "Service " + svc + " not found!"
                do_fail("check " + svc + " is running")
                continue
            act_name = get_name_by_pid(pid)
            if act_name == None:
                print "Service " + svc + " not actually running!"
                do_fail("check " + svc + " is running")                
            elif exe_name == None:
                print "Please add " + svc + " to DAEMON_NAMES!"
                do_skip("check " + svc + " is running")
            elif exe_name != act_name:
                print ("Pid #" + str(pid) + " is '" + act_name +
                       "', expected '" + exe_name + "'")
                do_fail("check " + svc + " is running")                
            else:
                do_pass("check " + svc + " is running")

def _main():
    test_slave.run_listener()
    check_what_up(EXPECT_STATUS)
    chk_really_runs()
    chk_ngc_updown()
    chk_service_updown()
    chk_daemon_pidfile_nofork()
    chk_daemon_pidfile_fork()
    chk_all_types("generic",lambda(x):True)
    chk_all_types("chdir",_chk_chdir_cb)
    chk_all_types("env",_chk_env_cb)
    try:
        u = pwd.getpwnam("nobody")
    except KeyError:
        u = None
    if u == None:
        do_log("No user 'nobody'")
        chk_all_types("suid",None)
    else:
        chk_all_types("suid",lambda(slave):_chk_suid_cb(u.pw_uid,slave))
    orig_stat = ngc_s()
    chk_ngc_c()
    check_what_up(orig_stat)
    summary()
    do_log("Finished!")

if __name__ == '__main__':
    _main()
