
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdio.h>
#include <string.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"

#include <assert.h>
#include <pwd.h>
#include <grp.h>

s_entry SUID = { "suid", STRING, NULL, "Set this user id on launch." };
s_entry SGID = { "sgid", STRING, NULL, "Set this group id on launch." };
static int do_suid(active_db_h * s, process_h * p __attribute__ ((unused)))
{
    const char *tmp = NULL;
    int ret = TRUE;

    assert(s);
    assert(s->name);
    assert(p);

    D_("do_suid!\n");

    if ((tmp = initng_active_db_get_string(&SGID, s)))
    {
        struct group *gr = getgrnam(tmp);

        if (!gr)
        {
            F_("%s: GROUP \"%s\" not found!\n", s->name, tmp);
            ret = FALSE;
        }
        D_("setgid to: %i.\n", gr->gr_gid);
        setgid(gr->gr_gid);
    }

    if ((tmp = initng_active_db_get_string(&SUID, s)))
    {
        struct passwd *pw = getpwnam(tmp);

        if (!pw)
        {
            F_("%s: USER \"%s\" not found!\n", s->name, tmp);
            ret = FALSE;
        }

        if (!initng_active_db_is(&SGID, s))
        {
            D_("setgid to: %i.\n", pw->pw_gid);
            setgid(pw->pw_gid);
        }

        initgroups(pw->pw_name, pw->pw_gid);

        D_("setuid to: %i.\n", pw->pw_uid);
        setuid(pw->pw_uid);
    }
    return ret;
}

int module_init(int api_version)
{
    D_("module_init();\n");
    if (api_version != API_VERSION)
    {
        F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", INITNG_VERSION, api_version);
        return (FALSE);
    }

    initng_service_data_types_add(&SUID);
    initng_service_data_types_add(&SGID);
    return (initng_plugin_hook_add(&g.A_FORK, 90, &do_suid));
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_service_data_types_del(&SUID);
    initng_service_data_types_del(&SGID);
    initng_plugin_hook_del(&g.A_FORK, &do_suid);
}
